package com.tzld.piaoquan.recommend.feature.service;

import com.google.common.cache.CacheBuilder;
import com.google.common.cache.CacheLoader;
import com.google.common.cache.LoadingCache;
import com.google.common.reflect.TypeToken;
import com.tzld.piaoquan.recommend.feature.util.CommonCollectionUtils;
import com.tzld.piaoquan.recommend.feature.util.JSONUtils;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.MapUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.dao.DataAccessException;
import org.springframework.data.redis.core.RedisOperations;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.data.redis.core.SessionCallback;
import org.springframework.data.redis.core.ValueOperations;

import java.util.*;
import java.util.concurrent.TimeUnit;
import java.util.function.Function;

/**
 * @author dyp
 */
@Slf4j
public abstract class AbstractFeatureService<K, V> {
    @Qualifier("redisTemplate")
    @Autowired
    private RedisTemplate<String, String> redisTemplate;

    @Qualifier("tairTemplate")
    @Autowired
    private RedisTemplate<String, String> tairTemplate;

    protected String emptyData;
    // 单位秒
    protected long emptyDataExpire;
    protected long defaultExpire;
    protected Function<K, String> keyFunc;
    protected TypeToken<V> typeToken;
    protected Function<K, V> defaultValueFunc;

    private int defaultMaximumSize = 10000;
    // 单位秒
    private int defaultRefreshAfterWrite = 60;
    private int defaultExpireAfterWrite = 60;
    private int defaultExpireAfterAccess = 60;
    private LoadingCache<K, V> cache;

    public AbstractFeatureService() {
        this.emptyData = "{}";
        // 单位秒
        this.emptyDataExpire = -1;
        this.defaultExpire = -1;
        this.keyFunc = k -> String.valueOf(k);
        this.defaultValueFunc = k -> null;

        initLocalCache(defaultMaximumSize, defaultRefreshAfterWrite, defaultExpireAfterWrite, defaultExpireAfterAccess);
    }

    protected void initLocalCache(int maximumSize, int refreshAfterWrite, int expireAfterWrite,
                                  int expireAfterAccess) {
        cache = CacheBuilder.newBuilder()
                .maximumSize(maximumSize)
                .refreshAfterWrite(refreshAfterWrite, TimeUnit.SECONDS)
                .expireAfterWrite(expireAfterWrite, TimeUnit.SECONDS)
                .expireAfterAccess(expireAfterAccess, TimeUnit.SECONDS)
                .build(new CacheLoader<K, V>() {
                    @Override
                    public V load(K key) {
                        return AbstractFeatureService.this.load(key);
                    }

                    @Override
                    public Map<K, V> loadAll(Iterable<? extends K> keys) {
                        return AbstractFeatureService.this.loadAll(keys);
                    }
                });
    }

    protected V get(K key) {
        assert cache != null;
        try {
            return cache.getUnchecked(key);
        } catch (Exception e) {
            log.error("get local cache error", e);
        }
        return null;
    }


    private V load(K key) {
        V feature = getFromCache(key);
        if (feature == null) {
            feature = defaultValueFunc.apply(key);
        }
//        if (feature == null) {
//            feature = getFromSource(key);
//            saveToCache(key, feature);
//        }
        return feature;
    }

    private V getFromCache(K key) {
        String redisKey = cacheKey(key);
        String value = redisTemplate.opsForValue().get(redisKey);
        return JSONUtils.fromJson(value, typeToken, null);
    }

    private V getFromSource(K key) {
        String sourceKey = cacheKey(key);
        String value = tairTemplate.opsForValue().get(sourceKey);
        return JSONUtils.fromJson(value, typeToken, null);
    }

    private void saveToCache(K key, V value) {
        String cacheKey = cacheKey(key);
        if (value == null) {
            redisTemplate.opsForValue().set(cacheKey, emptyData, emptyDataExpire, TimeUnit.SECONDS);
        } else {
            redisTemplate.opsForValue().set(cacheKey, JSONUtils.toJson(value), defaultExpire, TimeUnit.SECONDS);
        }
    }

    private Map<K, V> loadAll(Iterable<? extends K> keyIte) {

        Iterator<? extends K> ite = keyIte.iterator();
        List<K> keys = new ArrayList<>();
        while (ite.hasNext()) {
            keys.add(ite.next());
        }


        Map<K, V> cacheFeatureMap = getFromCache(keys);
        for (int i = 0; i < keys.size(); i++) {
            if (cacheFeatureMap.get(keys.get(i)) == null) {
                cacheFeatureMap.put(keys.get(i), defaultValueFunc.apply(keys.get(i)));
            }
        }
        return cacheFeatureMap;

//        if (cacheFeatureMap.size() == keys.size()) {
//            return cacheFeatureMap;
//        }
//        ite = keys.iterator();
//        while (ite.hasNext()) {
//            cacheFeatureMap.containsKey(ite.next());
//            ite.remove();
//        }
//
//        Map<K, V> sourceFeatureMap = null;
//        if (CollectionUtils.isNotEmpty(keys)) {
//            sourceFeatureMap = getFromSource(keys);
//            // TODO 可异步
//            saveToCache(keys, sourceFeatureMap);
//        }
//        return CommonCollectionUtils.merge(cacheFeatureMap, sourceFeatureMap);
    }

    private Map<K, V> getFromCache(List<K> keys) {
        List<String> redisKeys = CommonCollectionUtils.toList(keys, this::cacheKey);
        List<String> values = redisTemplate.opsForValue().multiGet(redisKeys);
        Map<K, V> result = new HashMap<>();
        for (int i = 0; i < keys.size(); i++) {
            if (values.get(i) != null) {
                result.put(keys.get(i), JSONUtils.fromJson(values.get(i), typeToken, null));
            }
        }
        return result;
    }

    private Map<K, V> getFromSource(List<K> keys) {
        List<String> redisKeys = CommonCollectionUtils.toList(keys, this::cacheKey);
        List<String> values = tairTemplate.opsForValue().multiGet(redisKeys);
        Map<K, V> result = new HashMap<>();
        for (int i = 0; i < keys.size(); i++) {
            if (values.get(i) != null) {
                result.put(keys.get(i), JSONUtils.fromJson(values.get(i), typeToken, null));
            }
        }
        return result;
    }

    private void saveToCache(List<K> keys, Map<K, V> map) {

        Map<String, String> emptyDataMap = new HashMap<>();
        Map<String, String> dataMap = new HashMap<>();
        for (K key : keys) {
            if (map.containsKey(key)) {
                dataMap.put(cacheKey(key), JSONUtils.toJson(map.get(key)));
            } else {
                emptyDataMap.put(cacheKey(key), emptyData);
            }
        }

        // TODO 评估过期时间
        redisTemplate.executePipelined(new SessionCallback<String>() {
            @Override
            public <A, B> String execute(RedisOperations<A, B> redisOperations) throws DataAccessException {
                ValueOperations<String, String> operations = (ValueOperations<String, String>) redisOperations.opsForValue();
                if (MapUtils.isNotEmpty(dataMap)) {
                    dataMap.entrySet().forEach(e ->
                            operations.set(e.getKey(), e.getValue(), defaultExpire, TimeUnit.SECONDS)
                    );
                }
                if (MapUtils.isNotEmpty(emptyDataMap)) {
                    emptyDataMap.entrySet().forEach(e ->
                            operations.set(e.getKey(), e.getValue(), emptyDataExpire, TimeUnit.SECONDS)
                    );
                }
                return null;
            }
        });

    }

    private String cacheKey(K key) {
        return keyFunc.apply(key);
    }

    protected void save(K key, V value) {
        saveToSource(key, value);
        // cache aside
        deleteFromCache(key);
    }

    private void saveToSource(K key, V value) {
        String sourceKey = cacheKey(key);
        String sourceValue = JSONUtils.toJson(value);
        tairTemplate.opsForValue().set(sourceKey, sourceValue);
    }

    private void deleteFromCache(K key) {
        String cacheKey = cacheKey(key);
        redisTemplate.delete(cacheKey);
    }

    protected Map<K, V> getAll(List<K> videoIds) {
        try {
            return cache.getAll(videoIds);
        } catch (Exception e) {
            log.error("get all failed {}", JSONUtils.toJson(videoIds), e);
        }
        return Collections.emptyMap();
    }

}
