package com.tzld.piaoquan.recommend.feature.service;

import com.alibaba.fastjson.JSONObject;
import com.tzld.piaoquan.recommend.feature.model.feature.GetUserFeatureInfo;
import com.tzld.piaoquan.recommend.feature.model.feature.GetVideoFeatureInfo;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.stereotype.Service;
import org.springframework.util.CollectionUtils;

import javax.annotation.Resource;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.*;
import java.util.stream.Collectors;

/**
 * @author sunxy
 */
@Service
public class UserAndVideoFeatureService {

    ExecutorService executorService = new ThreadPoolExecutor(16, 16, 0L,
            TimeUnit.MILLISECONDS, new LinkedBlockingQueue<>());

    private final static Logger log = LoggerFactory.getLogger(UserAndVideoFeatureService.class);
    private final static String USER_FEATURE_KEY = "recommend.feature.user.recsys.info.{mid}";

    private final static String VIDEO_FEATURE_KEY = "recommend.feature.video.recsys.info.{videoId}";

    @Resource(name = "redisTemplate")
    private RedisTemplate<String, Object> redisTemplate;

    public List<GetUserFeatureInfo> queryUserFeature(List<String> mids) {
        if (CollectionUtils.isEmpty(mids)) {
            return Collections.emptyList();
        }
        // 一次读取不能超过 2000 条
        if (mids.size() > 2000) {
            mids = mids.subList(0, 2000);
        }
        // 批量读取 Redis
        List<String> keys = mids.stream().map(mid -> USER_FEATURE_KEY.replace("{mid}", mid)).collect(Collectors.toList());
        List<Object> resultObjList = redisTemplate.opsForValue().multiGet(keys);
        if (CollectionUtils.isEmpty(resultObjList)) {
            return Collections.emptyList();
        }
        List<GetUserFeatureInfo> result = Collections.synchronizedList(new ArrayList<>(resultObjList.size()));
        // 提交 Callable 任务
        List<Callable<Void>> callableList = new ArrayList<>(resultObjList.size());
        for (Object resultObj : resultObjList) {
            callableList.add(() -> {
                String resultJsonStr = (String) resultObj;
                try {
                    JSONObject resultJson = JSONObject.parseObject(resultJsonStr);
                    if (resultJson == null) {
                        return null;
                    }
                    // convertToProto
                    GetUserFeatureInfo getUserFeatureInfo = UserAndVideoFeatureConverter.convertUserFeatureInfoToProto(resultJson);
                    result.add(getUserFeatureInfo);
                } catch (Exception e) {
                    log.error("queryUserFeature error, resultJson:{}", resultJsonStr, e);
                }
                return null;
            });
        }

        // 并发执行字段转换操作
        try {
            executorService.invokeAll(callableList);
        } catch (InterruptedException e) {
            log.error("queryUserFeature error, mids:{}", mids, e);
        }

        return result;
    }

    public List<GetVideoFeatureInfo> queryVideoFeature(List<Long> videoIds) {
        if (CollectionUtils.isEmpty(videoIds)) {
            return Collections.emptyList();
        }
        // 一次读取不能超过 2000 条
        if (videoIds.size() > 2000) {
            videoIds = videoIds.subList(0, 2000);
        }
        // 批量读取 Redis
        List<String> keys = videoIds.stream().map(videoId -> VIDEO_FEATURE_KEY.replace("{videoId}", String.valueOf(videoId))).collect(Collectors.toList());
        List<Object> resultObjList = redisTemplate.opsForValue().multiGet(keys);
        if (CollectionUtils.isEmpty(resultObjList)) {
            return Collections.emptyList();
        }
        // 提交 Callable 任务
        List<GetVideoFeatureInfo> result = Collections.synchronizedList(new ArrayList<>(resultObjList.size()));
        List<Callable<Void>> callableList = new ArrayList<>(resultObjList.size());
        for (Object resultObj : resultObjList) {
            callableList.add(() -> {
                String resultJsonStr = (String) resultObj;
                try {
                    JSONObject resultJson = JSONObject.parseObject(resultJsonStr);
                    if (resultJson == null) {
                        return null;
                    }
                    // convertToProto
                    GetVideoFeatureInfo getVideoFeatureInfo = UserAndVideoFeatureConverter.convertVideoFeatureInfoToProtoVideo(resultJson);
                    result.add(getVideoFeatureInfo);
                } catch (Exception e) {
                    log.error("queryVideoFeature error, resultJson:{}", resultJsonStr, e);
                }
                return null;
            });
        }

        // 并发执行字段转换操作
        try {
            executorService.invokeAll(callableList);
        } catch (InterruptedException e) {
            log.error("queryVideoFeature error, videoIds:{}", videoIds, e);
        }

        return result;
    }

}
